-- ********************************************************************************
-- Binarizer Testbench for HEVC h.265         
-- 01/03/2017                 
-- Norwegian University of Science and Technology              
-- Lars Erik Songe Paulsen             
-- ********************************************************************************

library ieee;
library std;
use std.textio.all;


use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_unsigned.all;

library work;
use work.BinarizerParameters.all;


entity BinarizerTestbench is

end BinarizerTestbench;

architecture struct of BinarizerTestbench is 


    -- ---------------------------------------------------------------------------
    -- Signal declarations
    -- ---------------------------------------------------------------------------
    
    signal Clk                      : std_logic;
    signal DataIn                   : std_logic_vector((16*CoeffWidth)-1 downto 0);
    signal ScanDir                  : std_logic_vector(1 downto 0);
    signal Resetn                   : std_logic;
    signal StartBinarizer           : std_logic;
    signal DataOut                  : std_logic_vector(OutputWidth-1 downto 0);
    signal DataLength               : std_logic_vector(OutputWidthLength-1 downto 0);
    signal PrefixLength             : std_logic_vector(OutputWidthLength-1 downto 0);
    signal Finished                 : std_logic;

    -- ---------------------------------------------------------------------------
    -- Test input
    -- ---------------------------------------------------------------------------

    --file TestFile : TEXT open read_mode is "C:\Users\lepaulse\Documents\GitHub\CABAC\HEVC_CABAC_Verification_Tool\HEVC_CABAC_Verification_Tool\bin\Debug\BinarizerTestfile.txt";
    file TestFile : TEXT open read_mode is "C:\Users\lepaulse\Desktop\CABACFiles\Binarizer_Table_Based_ALRem\BinarizerTestfile.txt";

    -- ---------------------------------------------------------------------------
    -- Functions
    -- ---------------------------------------------------------------------------
    
    -- Parser from TestFile input
    function string_to_binary(inp: string) return std_logic_vector is 
        variable temp: std_logic_vector(inp'length-1 downto 0) := (others => 'X'); 
        begin for i in inp'range loop 
            case inp(i) is 
                when '0' => temp(i-1) := '0'; 
                when '1' => temp(i-1) := '1'; 
                when others => temp(i-1) := 'X'; 
            end case; 
        end loop; 
        report "Parsed: " 
            & integer'image(to_integer(signed(temp(((CoeffWidth*16)-1) downto (CoeffWidth*15))))) & "," 
            & integer'image(to_integer(signed(temp(((CoeffWidth*15)-1) downto (CoeffWidth*14))))) & "," 
            & integer'image(to_integer(signed(temp(((CoeffWidth*14)-1) downto (CoeffWidth*13))))) & "," 
            & integer'image(to_integer(signed(temp(((CoeffWidth*13)-1) downto (CoeffWidth*12))))) & "," 
            & integer'image(to_integer(signed(temp(((CoeffWidth*12)-1) downto (CoeffWidth*11))))) & "," 
            & integer'image(to_integer(signed(temp(((CoeffWidth*11)-1) downto (CoeffWidth*10))))) & "," 
            & integer'image(to_integer(signed(temp(((CoeffWidth*10)-1) downto (CoeffWidth*9)))))  & "," 
            & integer'image(to_integer(signed(temp(((CoeffWidth*9)-1)  downto (CoeffWidth*8)))))  & "," 
            & integer'image(to_integer(signed(temp(((CoeffWidth*8)-1)  downto (CoeffWidth*7)))))  & "," 
            & integer'image(to_integer(signed(temp(((CoeffWidth*7)-1)  downto (CoeffWidth*6)))))  & "," 
            & integer'image(to_integer(signed(temp(((CoeffWidth*6)-1)  downto (CoeffWidth*5)))))  & "," 
            & integer'image(to_integer(signed(temp(((CoeffWidth*5)-1)  downto (CoeffWidth*4)))))  & "," 
            & integer'image(to_integer(signed(temp(((CoeffWidth*4)-1)  downto (CoeffWidth*3)))))  & "," 
            & integer'image(to_integer(signed(temp(((CoeffWidth*3)-1)  downto (CoeffWidth*2)))))  & "," 
            & integer'image(to_integer(signed(temp(((CoeffWidth*2)-1)  downto (CoeffWidth*1)))))  & "," 
            & integer'image(to_integer(signed(temp(((CoeffWidth*1)-1)  downto (CoeffWidth*0)))))  & "," 
            & "from TestFile";
        return temp; 
    end function string_to_binary;

    procedure FormatInput(
        signal   DataIn : out std_logic_vector((16*CoeffWidth)-1 downto 0)) is
        variable fileline : line;
        variable index  : integer range (CoeffWidth*16)-1 downto 0 := (CoeffWidth*16)-1;
    begin 
        for i in 0 to 15 loop
            readline(TestFile, fileline); 
            for j in fileline'range loop
                if (fileline(j) = '1') then
                    DataIn(index) <= '1'; 
                    index := index - 1;
                else
                    DataIn(index) <= '0';
                    index := index - 1;
                end if;
            end loop;
        end loop;
    end FormatInput;


    -- ---------------------------------------------------------------------------
    -- DUT component
    -- ---------------------------------------------------------------------------

    component Binarizer is
        port 
        (
        Clk                         : in std_logic;
        DataIn                      : in std_logic_vector((16*CoeffWidth)-1 downto 0);
        ScanDir                     : in std_logic_vector(1 downto 0);
        Resetn                      : in std_logic;
        StartBinarizer              : in std_logic;
        DataOut                     : out std_logic_vector(OutputWidth-1 downto 0); 
        DataLength                  : out std_logic_vector(OutputWidthLength-1 downto 0); 
        PrefixLength                : out std_logic_vector(OutputWidthLength-1 downto 0);
        Finished                    : out std_logic
        );
    end component Binarizer;

    
begin

    -- ---------------------------------------------------------------------------
    -- DUT port map
    -- ---------------------------------------------------------------------------
    BinarizerInstance : Binarizer
        port map
        (
        Clk                     => Clk,
        DataIn                  => DataIn,
        ScanDir                 => ScanDir,
        Resetn                  => Resetn,
        StartBinarizer          => StartBinarizer,
        DataOut                 => DataOut,
        DataLength              => DataLength,
        PrefixLength            => PrefixLength,
        Finished                => Finished
        );

    ClkGen : process
    begin
        Clk <= '1';
        wait for 5 ns;
        Clk <= '0';
        wait for 5 ns;
    end process;

    TestGen : process
        
    -- ---------------------------------------------------------------------------
    -- Variable declarations
    -- ---------------------------------------------------------------------------

    variable TestString : string((16*CoeffWidth) downto 1);
    variable TestLine : line;


    begin
        while not endfile(TestFile) loop
            --readline(TestFile, TestLine);
            --read(TestLine, TestString);
            --DataIn <= string_to_binary(TestString);
            FormatInput(DataIn); 
            Resetn <= '0';
            wait for 5 ns;
            Resetn <= '1';
            wait until Finished = '1';
        end loop;
        file_close(TestFile);
        wait;
    end process;

    
end struct;