-- ********************************************************************************
-- Bypass coder for HEVC h.265       
-- 01/03/2017                 
-- Norwegian University of Science and Technology              
-- Lars Erik Songe Paulsen             
-- ********************************************************************************

-- ********************************************************************************
-- TODO LIST:    
-- ********************************************************************************
-- BinCountInNALUnits           
-- ********************************************************************************

--pragma synthesis_off
  -- your simulation-only code
--pragma synthesis_on

library ieee;
library std;
use ieee.std_logic_1164.all;
use IEEE.STD_LOGIC_ARITH.ALL;
use ieee.numeric_std.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all; -- http://www1.pldworld.com/@xilinx/html/technote/tool/manual/15i_doc/fndtn/vhd/vhd10_3.htm
use std.textio.all;

library work;
use work.BypassParameters.all;

entity Bypass is

    port 	
    (
    Clk                         : in std_logic;
    --Testing with 16 bit long testdata(K=0 & N=77)
    DataIn                      : in std_logic_vector(InputWidth-1 downto 0);
    DataInLength                : in std_logic_vector(InputWidthLength-1 downto 0);
    ContextModel                : in std_logic_vector(1 downto 0);
    Resetn                      : in std_logic;
    StartBypass                 : in std_logic;
    DataOut                     : out std_logic_vector(OutputWidth-1 downto 0); 
    DataOutLength               : out std_logic_vector(OutputWidthLength-1 downto 0);
    
    Finished                    : out std_logic
    );

end Bypass;

architecture struct of Bypass is

    -- ---------------------------------------------------------------------------
    -- Signal declarations
    -- ---------------------------------------------------------------------------

    type BypassStateType is(
                    read_input, 
                    new_bit,
                    In_PutBit,                    
                    write_finished
                );

    signal BypassState : BypassStateType;
	
begin

    Bypass : process(Clk, DataIn, Resetn, StartBypass, ContextModel)
    
    -- ---------------------------------------------------------------------------
    -- Variable declarations
    -- ---------------------------------------------------------------------------
    
    -- Variables specified in subclause 9.2.4.4 in ISO/IEX 23008-2 : 201x(E)
    variable codILow : std_logic_vector(10 downto 0);--unsigned(10 downto 0);
    variable codIRange : std_logic_vector(8 downto 0);--unsigned(8 downto 0);
    variable bitsOutstanding : integer range 0 to OutputWidth-1; 
    variable firstBitFlag : std_logic;

    variable PutBit : std_logic;
    variable PutBitIndex : integer range OutputWidth-1 downto 0;

    variable bits : std_logic_vector(InputWidth-1 downto 0);
    variable bitsIndex : integer range 0 to InputWidth-1;
    variable bitsLength : std_logic_vector(InputWidthLength-1 downto 0);
	
    begin
        if Resetn = '0' then
            DataOut        <= (others => '0');
            DataOutLength    <= (others => '0');
            BypassState  <= read_input;
            Finished       <= '1';
			--dummy values:
            codILow                         := (others => '0');--0
            codIRange                       := "111111110"; --510
            firstBitFlag := '1';
            PutBitIndex := 0;
            bitsOutstanding := 0;
        elsif rising_edge(Clk) then
            case BypassState is
                when read_input =>
                    if (StartBypass = '1') then
                        BypassState <= new_bit;
                        bits        := DataIn;
                        bitsLength  := DataInLength;
                        Finished    <= '0';
                        bitsIndex   := InputWidth-1;
                        PutBitIndex := 0;
                    end if;
                when new_bit => 
                    if (bitsIndex>=(InputWidth-conv_integer(bitsLength))) then
                        codILow := codILow(9 downto 0) & "0";
                        if (bits(bitsIndex) /= '0') then
                        codILow := codILow + codIRange;
                        end if;
                        if(codILow>=1024) then
                            PutBit := '1';
                            BypassState <= In_putBit;
                            codILow := codILow - 1024;
                        elsif (codILow<512) then
                            PutBit := '0';
                            BypassState <= In_putBit;
                        else
                            codILow := codILow - 512;
                            bitsOutstanding := bitsOutstanding + 1;
                        end if;
                        bitsIndex := bitsIndex - 1;
                    else
                        DataOutLength <= std_logic_vector(to_unsigned(PutBitIndex,OutputWidthLength));
                        BypassState <= write_finished;                                
                    end if;
                when In_PutBit =>
                    if (firstBitFlag /= '0') then 
                        firstBitFlag := '0';
                    else
                        DataOut((OutputWidth-1)-PutBitIndex) <= PutBit;
                        PutBitIndex := PutBitIndex + 1;
                    end if;
                    PutBit_loop : for i in 0 to PutBitLoopLength-1 loop -- TODO: Potential overflow here if bitsOutstanding > PutBitLoopLength
                        if (bitsOutstanding > 0) then
                            DataOut((OutputWidth-1)-PutBitIndex) <= not PutBit;
                            bitsOutstanding := bitsOutstanding - 1;
                            PutBitIndex := PutBitIndex + 1;
                        else
                            BypassState <= new_bit;
                            exit PutBit_loop;
                        end if;
                    end loop;
				when others =>
                    --BypassState <= read_input;
                    BypassState <= read_input;
                    Finished <= '1';
                    DataOutLength <= std_logic_vector(to_unsigned(PutBitIndex,OutputWidthLength));
            end case;
        end if;
    end process;
end struct;