-- ********************************************************************************
-- Regular coder for HEVC h.265       
-- 01/03/2017                 
-- Norwegian University of Science and Technology              
-- Lars Erik Songe Paulsen             
-- ********************************************************************************

-- ********************************************************************************
-- TODO LIST:    
-- ********************************************************************************
--             
-- ********************************************************************************

--pragma synthesis_off
  -- your simulation-only code
--pragma synthesis_on

library ieee;
library std;
use ieee.std_logic_1164.all;
use IEEE.STD_LOGIC_ARITH.ALL;
use ieee.numeric_std.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all; -- http://www1.pldworld.com/@xilinx/html/technote/tool/manual/15i_doc/fndtn/vhd/vhd10_3.htm
use std.textio.all;
use ieee.std_logic_textio.all;

library work;
use work.RegularParameters.all;

entity Regular is

    port 	
    (
    Clk                         : in std_logic;
    DataIn                      : in std_logic_vector(InputWidth-1 downto 0);
    DataInLength                : in std_logic_vector(InputWidthLength-1 downto 0);
    ContextModel                : in std_logic_vector(1 downto 0);
    SliceQPY                    : in std_logic_vector(5 downto 0);
    initType                    : in std_logic_vector(1 downto 0);
    Resetn                      : in std_logic;
    StartRegular                : in std_logic;
    DataOut                     : out std_logic_vector(OutputWidth-1 downto 0); 
    DataOutLength               : out std_logic_vector(OutputWidthLength-1 downto 0);
    Finished                    : out std_logic
    );

end Regular;



architecture struct of Regular is

    signal qCodIRangeIdx : std_logic_VECTOR(0 to 1);
    signal pStateIdx : std_logic_VECTOR(0 to 5);
    -- Wires from Memmory interface
    signal rangeTabLPS_w : std_logic_vector(rangeTabLPS_precision-1 downto 0);
    signal transIdxLPS_w : std_logic_vector(rangeTabLPS_precision-1 downto 0);
    signal transIdxMPS_w : std_logic_vector(rangeTabLPS_precision-1 downto 0);

    -- ---------------------------------------------------------------------------
    -- Table declarations
    -- ---------------------------------------------------------------------------

    type rom_64 is array (0 to 63) of std_logic_vector (rangeTabLPS_precision-1 downto 0);
    type rom_4x64 is array (0 to  3) of rom_64;


    impure function InitRomFromFile (RomFileName : in string) return rom_64 is
        FILE romfile : text is in RomFileName;
        variable RomFileLine : line;
        variable rom : rom_64;
    begin
        for i in rom_64'range loop
            readline(romfile, RomFileLine);
            hread(RomFileLine, rom(i));
        end loop;
        return rom;
    end function;

    signal rangeTabLPS : rom_4x64   := (InitRomFromFile("qCodIRangeIdx0.txt"), 
                                        InitRomFromFile("qCodIRangeIdx1.txt"), 
                                        InitRomFromFile("qCodIRangeIdx2.txt"),
                                        InitRomFromFile("qCodIRangeIdx3.txt"));
    signal transIdxLPS : rom_64     :=  InitRomFromFile("transIdxLPS.txt");
    signal transIdxMPS : rom_64     :=  InitRomFromFile("transIdxMPS.txt");





    type ctxTbl is array (0 to 17939) of std_logic_vector(7 downto 0); -- 1bit for Initialized, 1bit for MPS and 6bits for probability 


    function string_to_binary(inp: string) return std_logic_vector is 
        variable temp: std_logic_vector(inp'length-1 downto 0) := (others => 'X'); 
        --report "input is: " & integer'image(inp'length);
        begin for i in inp'range loop 
            case inp(i) is 
                when '0' => temp(i-1) := '0'; 
                when '1' => temp(i-1) := '1'; 
                when others => temp(i-1) := 'X'; 
            end case; 
        end loop;
        return temp; 
    end function string_to_binary;

    impure function InitctxTblFromFile (RomFileName : in string) return ctxTbl is
        FILE romfile : text is in RomFileName;
        variable RomFileLine : line;
        variable rom : ctxTbl;
        variable TestString : string(8 downto 1);
    begin
        for i in ctxTbl'range loop
            readline(romfile, RomFileLine);
            read(RomFileLine, TestString);
            rom(i) := string_to_binary(TestString);
        end loop;
        return rom;
    end function;

    -- 9.2.1.1  Initialization process for context variables
    signal ctxIdxTable : ctxTbl     := InitctxTblFromFile("Output.txt");




    -- ---------------------------------------------------------------------------
    -- Signal declarations
    -- ---------------------------------------------------------------------------

    type RegularStateType is(
                    read_input, 
                    code_bits,
                    write_finished
                );

    signal RegularState : RegularStateType;
	
begin

    -- ---------------------------------------------------------------------------
    -- Memory interfacing
    -- ---------------------------------------------------------------------------

    --TableLookups : process(Clk)
    --begin
    --    if rising_edge(Clk) then
    --        rangeTabLPS_w <= rangeTabLPS(conv_integer(qCodIRangeIdx))(conv_integer(pStateIdx));
    --        transIdxLPS_w <= transIdxLPS(conv_integer(pStateIdx));
    --        transIdxMPS_w <= transIdxMPS(conv_integer(pStateIdx));
    --    end if;
    --end process;


    Regular : process(Clk, DataIn, Resetn, StartRegular, ContextModel)
    
    -- ---------------------------------------------------------------------------
    -- Variable declarations
    -- ---------------------------------------------------------------------------
    
    -- Variables specified in subclause 9.2.4.4 in ISO/IEX 23008-2 : 201x(E)
    --variable codILow : integer range 0 to 1023;--unsigned(10 downto 0);
    --variable codIRange : integer range 0 to 511;--unsigned(9 downto 0);

    variable codILow : std_logic_vector(10 downto 0);
    variable codIRange : std_logic_vector(9 downto 0);

    variable codIRangeLPS : std_logic_vector(rangeTabLPS_precision-1 downto 0);
    --variable transIdxLPS : std_logic_vector(rangeTabLPS_precision-1 downto 0);
    --variable transIdxMPS : std_logic_vector(rangeTabLPS_precision-1 downto 0);

    variable bitsOutstanding : integer range 0 to InputWidth-1; 
    variable firstBitFlag : std_logic;

    variable binVal_index : integer range 0 to InputWidth-1;
    variable PutBit_index : integer range 0 to InputWidth-1;
    --variable codI : std_logic_vector(InputWidth-1 downto 0);

    --variable bins : std_logic_vector(InputWidth-1 downto 0);
	
    begin
        if Resetn = '0' then
            DataOut        <= (others => '0');
            DataOutLength    <= (others => '0');
            RegularState  <= read_input;
            Finished       <= '1';
            qCodIRangeIdx <= "00";
            pStateIdx       <= "000000";
			--dummy values:
            codILow := (others => '0');
            codIRange := "0111111110";
            binVal_index := 15;
            firstBitFlag := '1';
            PutBit_index := 7;
            bitsOutstanding := 0;
        elsif rising_edge(Clk) then
            case RegularState is
                when read_input =>
                    RegularState <= code_bits;
                    Finished      <= '0';
                    --bins := DataIn;
                when code_bits =>   
                    --RegularState <= code_bits;
                    RegularState <= write_finished;
                    code_bits_loop : for i in InputWidth-1 downto 0 loop
                        qCodIRangeIdx <= codIRange(7 downto 6);
                        -- PLACEHOLDER TABLE LOOKUPS
                        codIRangeLPS := rangeTabLPS(conv_integer(qCodIRangeIdx))
                                                   (conv_integer(ctxIdxTable((conv_integer(SliceQPY)*345)+(conv_integer(initType)*115))(5 downto 0)));
                        codIRange := codIRange - codIRangeLPS;
                        if(ctxIdxTable((conv_integer(SliceQPY)*345)+(conv_integer(initType)*115))(6) /= DataIn(i)) then
                            codILow := codILow + codIRange;
                            codIRange := "00" & codIRangeLPS;
                            if(conv_integer(ctxIdxTable((conv_integer(SliceQPY)*345)+(conv_integer(initType)*115))(5 downto 0)) /= 0) then 
                                ctxIdxTable((conv_integer(SliceQPY)*345)+(conv_integer(initType)*115))(6) <= not ctxIdxTable((conv_integer(SliceQPY)*345)+(conv_integer(initType)*115))(6);
                            end if;
                            ctxIdxTable((conv_integer(SliceQPY)*345)+(conv_integer(initType)*115))(5 downto 0) <= transIdxLPS(conv_integer(ctxIdxTable((conv_integer(SliceQPY)*345)+(conv_integer(initType)*115))(5 downto 0)))(5 downto 0);
                        else
                            ctxIdxTable((conv_integer(SliceQPY)*345)+(conv_integer(initType)*115))(5 downto 0) <= transIdxMPS(conv_integer(ctxIdxTable((conv_integer(SliceQPY)*345)+(conv_integer(initType)*115))(5 downto 0)))(5 downto 0);
                        end if;

                        if (codIRange < 256) then
                            if (codILow < 256) then
                                --PUTBIT0
                                if (firstBitFlag = '1') then 
                                firstBitFlag := '0';
                                else
                                    DataOut(PutBit_index) <= '0';
                                    PutBit_index := PutBit_index - 1;
                                end if;
                                put_bit_0_bitsOutstanding_loop : for j in InputWidth-1 downto 0 loop
                                    if (bitsOutstanding > 0) then
                                        DataOut(PutBit_index) <= '1';
                                        PutBit_index := PutBit_index - 1;
                                        bitsOutstanding := bitsOutstanding - 1;
                                    end if;
                                end loop;  
                            elsif(codILow >= 512) then
                                codIlow := codILow -512;
                                --PUTBIT1
                                if (firstBitFlag = '1') then 
                                    firstBitFlag := '0';
                                else
                                    DataOut(PutBit_index) <= '1';
                                    PutBit_index := PutBit_index - 1;
                                end if;
                                put_bit_1_bitsOutstanding_loop : for j in InputWidth-1 downto 0 loop
                                    if (bitsOutstanding > 0) then
                                        DataOut(PutBit_index) <= '0';
                                        PutBit_index := PutBit_index - 1;
                                        bitsOutstanding := bitsOutstanding - 1;
                                    end if;
                                end loop;  
                            else
                                codILow := codILow - 256;
                                bitsOutstanding := bitsOutstanding + 1;
                                report " bitsOutstanding := bitsOutstanding + 1 @ " & integer'image(i);
                            end if;
                            codIRange := codIRange(8 downto 0) & "0";
                            codILow := codILow(9 downto 0) & "0";
                        end if;
                    end loop;
                    report "codILow: " & integer'image(conv_integer(codILow));
                    report " bitsOutstanding when finished: " & integer'image(bitsOutstanding);
				when others =>
                    RegularState <= write_finished;
                    Finished <= '1';
                    DataOut        <= (others => '0');
                    DataOutLength    <= (others => '0');
            end case;
        end if;
    end process;
end struct;