-- ********************************************************************************
-- Regular coder Testbench for HEVC h.265	      
-- 01/03/2017			      
-- Norwegian University of Science and Technology              
-- Lars Erik Songe Paulsen             
-- ********************************************************************************

library ieee;
library std;
use std.textio.all;


use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_unsigned.all;

library work;
use work.RegularParameters.all;


entity RegularTestbench is

end RegularTestbench;

architecture struct of RegularTestbench is 


	-- ---------------------------------------------------------------------------
	-- Signal declarations
	-- ---------------------------------------------------------------------------
	
	signal Clk			            : std_logic;
	signal DataIn			        : std_logic_vector(InputWidth-1 downto 0);
    signal DataInLength             : std_logic_vector(InputWidthLength-1 downto 0);
	signal ContextModel			    : std_logic_vector(1 downto 0);
    signal SliceQPY                 : std_logic_vector(5 downto 0);
    signal initType                    : std_logic_vector(1 downto 0);
	signal Resetn	        		: std_logic;
	signal StartRegular   		    : std_logic;
	signal DataOut                	: std_logic_vector(OutputWidth-1 downto 0);
	signal DataOutLength            : std_logic_vector(OutputWidthLength-1 downto 0);
    signal Finished                 : std_logic;

	-- ---------------------------------------------------------------------------
	-- Test input
	-- ---------------------------------------------------------------------------

    -- Include TestFile.txt in Vivado simulation sources.
	file TestFile : TEXT open read_mode is "TestFile.txt";

	
	-- ---------------------------------------------------------------------------
	-- Functions
	-- ---------------------------------------------------------------------------
	
	-- Parser from TestFile input
	function string_to_binary(inp: string) return std_logic_vector is 
		variable temp: std_logic_vector(inp'length-1 downto 0) := (others => 'X'); 
        --report "input is: " & integer'image(inp'length);
		begin for i in inp'range loop 
			case inp(i) is 
				when '0' => temp(i-1) := '0'; 
				when '1' => temp(i-1) := '1'; 
				when others => temp(i-1) := 'X'; 
			end case; 
		end loop;
		return temp; 
	end function string_to_binary;


	-- ---------------------------------------------------------------------------
	-- DUT component
	-- ---------------------------------------------------------------------------

	component Regular is
		port 
		(
        Clk      		    		: in std_logic;
        DataIn      			    : in std_logic_vector(InputWidth-1 downto 0);
        DataInLength                : in std_logic_vector(InputWidthLength-1 downto 0);
        ContextModel 	    	    : in std_logic_vector(1 downto 0);
        SliceQPY                    : in std_logic_vector(5 downto 0);
        initType                    : in std_logic_vector(1 downto 0);
        Resetn			        	: in std_logic;
        StartRegular		    	: in std_logic;
        DataOut                	 	: out std_logic_vector(OutputWidth-1 downto 0); 
        DataOutLength	            : out std_logic_vector(OutputWidthLength-1 downto 0); 
        Finished                    : out std_logic
        );
	end component Regular;

	
begin

	-- ---------------------------------------------------------------------------
	-- DUT port map
	-- ---------------------------------------------------------------------------
	RegularInstance : Regular
		port map
		(
		Clk      				=> Clk,
		DataIn 			        => DataIn,
        DataInLength            => DataInLength,
		ContextModel			=> ContextModel,
        SliceQPY                => SliceQPY,
        initType                => initType,
		Resetn				    => Resetn,
		StartRegular			=> StartRegular,
		DataOut		            => DataOut,
		DataOutLength	        => DataOutLength,
        Finished                => Finished
		);

	ClkGen : process
	begin
		Clk <= '1';
		wait for 5 ns;
		Clk <= '0';
		wait for 5 ns;
	end process;

	TestGen : process
		
	-- ---------------------------------------------------------------------------
	-- Variable declarations
	-- ---------------------------------------------------------------------------

    --The status of the arithmetic decoding engine is represented by the variables codIRange and codIOffset. In the
    --initialization procedure of the arithmetic decoding process, codIRange is set equal to 510 and codIOffset is set equal to
    --the value returned from read_bits( 9 ) interpreted as a 9 bit binary representation of an unsigned integer with most
    --significant bit written first.

	variable TestString : string(InputWidth downto 1);
	variable TestLine : line;

    variable codIOffset : integer range 0 to 1023 := 0;
    variable codIRange : integer range 0 to 511 := 510;
    variable codIRangeVector : std_logic_vector(8 downto 0);
	begin
        initType <= "01";
        SliceQPY <= "000010";
		while not endfile(TestFile) loop
			readline(TestFile, TestLine);
			read(TestLine, TestString);
			DataIn <= string_to_binary(TestString);
			Resetn <= '0';
			wait for 5 ns;
			Resetn <= '1';
            wait for 6 ns;
            wait until Finished = '1';
                for i in 15 downto 0 loop

                end loop;
        end loop;
		wait;
	end process;
	
end struct;

