library ieee;
library std;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_unsigned.all;


entity CABAC is

	port (
        GCLK : in std_logic;
        LD : out std_logic_vector(7 downto 0);
        BTN : in std_logic_vector(4 downto 0);
        
        TestON : in std_logic;
        TestRESETN : in std_logic;    
        TestWREN : in std_logic;
        
        WORKING : inout std_logic;
        
        index_o : out std_logic_vector(3 downto 0);
        Testout1 : out std_logic_vector(31 downto 0);
        Testout2 : out std_logic_vector(31 downto 0);
        Coeffsout : out std_logic_vector(111 downto 0);
        
        Testin0 : in std_logic_vector(31 downto 0);
        Testin1 : in std_logic_vector(31 downto 0);
        Testin2 : in std_logic_vector(31 downto 0);
        Coeffsin : in std_logic_vector(111 downto 0)
	);
end CABAC;

architecture arch_imp of CABAC is

    signal counter : std_logic_vector(31 downto 0) := "00000000000000000000000000000000";
    signal ldcounter : std_logic_vector(3 downto 0) := "0000";
    
	signal index : unsigned(3 downto 0) := "0000";


begin


    CLOCKED: process(GCLK, BTN, Testin0, Testin1, Testin2, Coeffsin, TestON, TestRESETN, TestWREN)
    variable COND : boolean := true;

    type TRANSFORM_BLOCK_COEFFICIENTS is array (3 downto 0, 3 downto 0) of std_logic_vector(6 downto 0);
    variable TRANSFORM_BLOCK : TRANSFORM_BLOCK_COEFFICIENTS;
	
	-- type TRANSFORM_BLOCK_DIRECTIONS is array (15 downto 0) of std_logic_vector(1 downto 0);
	-- variable DIAGONAL_SCAN_X : TRANSFORM_BLOCK_DIRECTIONS := ("11", "11", "10", "11", "10", "01", "11", "10", "01", "00", "10", "01", "00", "01", "00", "00");
	-- variable DIAGONAL_SCAN_Y : TRANSFORM_BLOCK_DIRECTIONS := ("11", "10", "11", "01", "10", "11", "00", "01", "10", "11", "00", "01", "10", "00", "01", "00");
	
	-- type TRANSFORM_BLOCK_DIRECTIONS is array (3 downto 0, 3 downto 0) of integer range 0 to 3;
	
	-- constant DIAGONAL_SCAN_X : TRANSFORM_BLOCK_DIRECTIONS := ((3, 3, 2, 3), (2, 1, 3, 2), (1, 0, 2, 1), (0, 1, 0, 0));
	-- constant DIAGONAL_SCAN_Y : TRANSFORM_BLOCK_DIRECTIONS := ((3, 2, 3, 1), (2, 3, 0, 1), (2, 3, 0, 1), (2, 0, 1, 0));
	
		
	type TRANSFORM_BLOCK_DIRECTIONS is array (15 downto 0) of std_logic_vector(2 downto 0);
	
	constant DIAGONAL_SCAN_X_TRk : TRANSFORM_BLOCK_DIRECTIONS := ("001", "101", "100", "101", "100", "001", "101", "100", "001", "000", "100", "001", "000", "001", "000", "000");
	constant DIAGONAL_SCAN_Y_TRk : TRANSFORM_BLOCK_DIRECTIONS := ("101", "100", "101", "001", "100", "101", "000", "001", "100", "101", "000", "001", "100", "000", "001", "000");
	
		type TRANSFORM_BLOCK_DIRECTIONS_LENGTHS is array (15 downto 0) of integer range 0 to 2;
	
	constant DIAGONAL_SCAN_X_LENGTH : TRANSFORM_BLOCK_DIRECTIONS_LENGTHS := (1, 2, 2, 2, 2, 1, 2, 2, 1, 1, 2, 1, 1, 1, 1, 1);
	constant DIAGONAL_SCAN_Y_LENGTH : TRANSFORM_BLOCK_DIRECTIONS_LENGTHS := (2, 2, 2, 1, 2, 2, 1, 1, 2, 2, 1, 1, 2, 1, 1, 1);
	
	type TRANSFORM_BLOCK_REMAINING is array (3 downto 0, 3 downto 0) of integer range 0 to 15;
	constant REMAINING_TRANSFORM_BLOCK_LENGTHS : TRANSFORM_BLOCK_REMAINING := ((15, 13, 10, 6), (14, 11, 7, 3), (12, 8, 4, 1), (9, 5, 2, 0));
	-- 0	2	5	9
	-- 1	4	8	12
	-- 3	7	11	14
	-- 6	10	13	15

	
	variable i : integer range 0 to 15;
	variable j : integer range 0 to 15;
	variable k : integer range 0 to 127;
	
	variable bit_string_index : integer range 0 to 111 := 111;
	

    
    --variable temp : std_logic_vector(31 downto 0) := "00000000000000000000000000000000";
	
	--Condition vars:
	
	--flags with statuses:
	variable last_sig_coeff_found : boolean;
	variable last_sig_coeff_x : std_logic_vector(4 downto 0);
	variable last_sig_coeff_y : std_logic_vector(4 downto 0);
	
	
	variable sig_coeff_remaining_length : integer range 0 to 15;
	variable sig_coeff_index: integer range 0 to 15;
	variable sig_coeff_write_index : integer range 0 to 15;
	variable sig_coeff_flag : std_logic_vector(0 to 15); -- first location inferred from last_sig_coeff_x/y flags.
	variable sig_coeff_flag_to_output : std_logic_vector(0 to 15);

	variable sig_abs_level_greater1_flag : std_logic_vector(0 to 15); -- first location inferred from last_sig_coeff_x/y flags.
	variable sig_abs_level_greater1_flag_to_output : std_logic_vector(0 to 15);
	variable sig_abs_level_greater1_index: integer range 0 to 15;
	variable sig_abs_level_greater1_length: integer range 0 to 15;

	variable sig_abs_level_greater2_flag : std_logic_vector(0 to 15); -- first location inferred from last_sig_coeff_x/y flags.
	
	type Binarization_Steps is (last_sig_coeff, sig_coeff, coeff_abs_level_greater1);
	
    begin
    
	
    if TestON = '1' then

		if rising_edge(GCLK) then
			if TestRESETN = '0' then
				counter <= (others => '0');
				index <= "1111";
				Testout1 <= (others => '0');
				Testout2 <= (others => '0');
				
				Coeffsout <= (others => '0');
				
				WORKING <= '0';
				
				index_o <= "1111";
				
				last_sig_coeff_found := FALSE;
				
				sig_coeff_remaining_length := 15;
				sig_coeff_index := 0;
				sig_coeff_write_index := 0;
				sig_coeff_flag := (others => '0');
				sig_coeff_flag_to_output := (others => '0');
				bit_string_index := 111;
				
			else
			
				if WORKING = '1' then
				
					--report "sig_coeff_flag: " & std_logic'image(sig_coeff_flag(0)) & std_logic'image(sig_coeff_flag(1)) & std_logic'image(sig_coeff_flag(2)) & std_logic'image(sig_coeff_flag(3)) & std_logic'image(sig_coeff_flag(4)) & std_logic'image(sig_coeff_flag(5)) & std_logic'image(sig_coeff_flag(6)) & std_logic'image(sig_coeff_flag(7)) & std_logic'image(sig_coeff_flag(8)) & std_logic'image(sig_coeff_flag(9)) & std_logic'image(sig_coeff_flag(10)) & std_logic'image(sig_coeff_flag(11)) & std_logic'image(sig_coeff_flag(12)) & std_logic'image(sig_coeff_flag(13)) & std_logic'image(sig_coeff_flag(14)) & std_logic'image(sig_coeff_flag(15));
					
					for bit_builder in last_sig_coeff to coeff_abs_level_greater1 loop
					
						case bit_builder is
							when last_sig_coeff =>
								-- Find sig_coeff_remaining_length loop
								Find_sig_coeff_remaining_length_loop:for sig_coeff_index in 0 to 15 loop
								
									if(sig_coeff_flag(sig_coeff_index) = '1') then
									
										sig_coeff_remaining_length := 15 - sig_coeff_index;
										sig_coeff_write_index := sig_coeff_index;
										--sig_coeff_flag := sig_coeff_flag sll sig_coeff_index;
										
										case sig_coeff_index is
										
											-- last_sig_coeff_x (101, 101, 100, 101, 100, 01, 101, 100, 01, 00, 100, 01, 00, 01, 00, 00)
											-- last_sig_coeff_y (101, 100, 101, 01, 100, 101, 00, 01, 100, 101, 00, 01, 100, 00, 01, 00)
											
																					--Truncated Rice
																					--TODO: MAJOR fix shifting	
											when 0 => 	Coeffsout(111 downto 106) 	<= "101101"; 	bit_string_index := 105;	sig_coeff_flag_to_output(0 to 14) 		:= sig_coeff_flag(1 to 15);		report  "sig_coeff_index is 0";
											when 1 => 	Coeffsout(111 downto 106) 	<= "101100"; 	bit_string_index := 105;	sig_coeff_flag_to_output(0 to 13) 		:= sig_coeff_flag(2 to 15);		report  "sig_coeff_index is 1";
											when 2 => 	Coeffsout(111 downto 106) 	<= "100101"; 	bit_string_index := 105;	sig_coeff_flag_to_output(0 to 12) 		:= sig_coeff_flag(3 to 15);		report  "sig_coeff_index is 2";
											when 3 => 	Coeffsout(111 downto 107) 	<= "10101"; 	bit_string_index := 106;	sig_coeff_flag_to_output(0 to 11) 		:= sig_coeff_flag(4 to 15);		report  "sig_coeff_index is 2"; --sig_coeff_flag 	:= sig_coeff_flag(4 to 15) & "0000";		report  "sig_coeff_index is 3"; -- zeroing out after bit shiting maybe redundant.
											when 4 => 	Coeffsout(111 downto 106) 	<= "100100"; 	bit_string_index := 105;	sig_coeff_flag_to_output(0 to 10) 		:= sig_coeff_flag(5 to 15);		report  "sig_coeff_index is 4";
											when 5 => 	Coeffsout(111 downto 107) 	<= "01101"; 	bit_string_index := 106;	sig_coeff_flag_to_output(0 to 9) 		:= sig_coeff_flag(6 to 15);		report  "sig_coeff_index is 5";
											when 6 => 	Coeffsout(111 downto 107) 	<= "10100"; 	bit_string_index := 106;	sig_coeff_flag_to_output(0 to 8) 		:= sig_coeff_flag(7 to 15);		report  "sig_coeff_index is 6";
											when 7 => 	Coeffsout(111 downto 107) 	<= "10001"; 	bit_string_index := 106;	sig_coeff_flag_to_output(0 to 7) 		:= sig_coeff_flag(8 to 15);		report  "sig_coeff_index is 7";
											when 8 => 	Coeffsout(111 downto 107) 	<= "01100"; 	bit_string_index := 106;	sig_coeff_flag_to_output(0 to 6) 		:= sig_coeff_flag(9 to 15);		report  "sig_coeff_index is 8";
											when 9 => 	Coeffsout(111 downto 107) 	<= "00101"; 	bit_string_index := 106;	sig_coeff_flag_to_output(0 to 5) 		:= sig_coeff_flag(10 to 15);	report  "sig_coeff_index is 9";
											when 10 => 	Coeffsout(111 downto 107) 	<= "10000"; 	bit_string_index := 106;	sig_coeff_flag_to_output(0 to 4) 		:= sig_coeff_flag(11 to 15);	report  "sig_coeff_index is 10";
											when 11 => 	Coeffsout(111 downto 108) 	<= "0101"; 		bit_string_index := 107;	sig_coeff_flag_to_output(0 to 3) 		:= sig_coeff_flag(12 to 15);	report  "sig_coeff_index is 11";
											when 12 => 	Coeffsout(111 downto 107) 	<= "00100"; 	bit_string_index := 106;	sig_coeff_flag_to_output(0 to 2) 		:= sig_coeff_flag(13 to 15);	report  "sig_coeff_index is 12";
											when 13 => 	Coeffsout(111 downto 108) 	<= "0100"; 		bit_string_index := 107;	sig_coeff_flag_to_output(0 to 1) 		:= sig_coeff_flag(14 to 15);	report  "sig_coeff_index is 13";
											when 14 => 	Coeffsout(111 downto 108) 	<= "0001"; 		bit_string_index := 107;	sig_coeff_flag_to_output(0 to 0) 		:= sig_coeff_flag(15 to 15);	report  "sig_coeff_index is 14";
											when 15 => 	Coeffsout(111 downto 108) 	<= "0000"; 		bit_string_index := 107;	sig_coeff_flag_to_output  				:= (others => '0');				report  "sig_coeff_index is 15";
											
										end case;
										
										--report  "sig_coeff_remaining_length : " & integer'image(sig_coeff_remaining_length) & " sig_coeff_index before exiting: " & integer'image(sig_coeff_index) & " bit_string_index: " & integer'image(bit_string_index);
										
										exit Find_sig_coeff_remaining_length_loop;
										
									end if;
								
								end loop; 	
							-- Find sig_coeff_remaining_length loop
							
							when sig_coeff =>

								Coeffsout(bit_string_index downto bit_string_index - 15) <= sig_coeff_flag_to_output;
								bit_string_index := bit_string_index - sig_coeff_remaining_length;
								--report  "Coeffsout(bit_string_index downto bit_string_index - 15) <= sig_coeff_flag(15 downto 0);" & "sig_coeff_flag: " & std_logic'image(sig_coeff_flag(0)) & std_logic'image(sig_coeff_flag(1)) & std_logic'image(sig_coeff_flag(2)) & std_logic'image(sig_coeff_flag(3)) & std_logic'image(sig_coeff_flag(4)) & std_logic'image(sig_coeff_flag(5)) & std_logic'image(sig_coeff_flag(6)) & std_logic'image(sig_coeff_flag(7)) & std_logic'image(sig_coeff_flag(8)) & std_logic'image(sig_coeff_flag(9)) & std_logic'image(sig_coeff_flag(10)) & std_logic'image(sig_coeff_flag(11)) & std_logic'image(sig_coeff_flag(12)) & std_logic'image(sig_coeff_flag(13)) & std_logic'image(sig_coeff_flag(14)) & std_logic'image(sig_coeff_flag(15));
								
								for sig_abs_level_greater1_index in 0 to 15 loop
									if (sig_coeff_flag(sig_abs_level_greater1_index) = '1') then
										if (sig_abs_level_greater1_flag(sig_abs_level_greater1_index) = '1') then
											sig_abs_level_greater1_flag_to_output(sig_abs_level_greater1_index) := '1';
										else
											sig_abs_level_greater1_flag_to_output(sig_abs_level_greater1_index) := '0';
										end if;
										sig_abs_level_greater1_length := sig_abs_level_greater1_length + 1;
									end if;
								end loop;

								
								
							when coeff_abs_level_greater1 =>
								
								Coeffsout(bit_string_index downto bit_string_index - 15) <= sig_coeff_flag_to_output;
								bit_string_index := bit_string_index - sig_abs_level_greater1_length;
		
						end case;
						
					
					
					end loop;
					
						WORKING <= '0';
					
					
				elsif TestWREN = '1' then -- Write to logic here
					
					WORKING <= '1'; -- Start working
					
					-- Testout1(31 downto 25) <= TRANSFORM_BLOCK(3,3);
					
					--Coeffsout <= (others => '0');							--reset Coeffsout
					
					-- Test vector 1
					-- 0001111	0001101	0001010	0000110	
					-- 0001110	0001011	0000111	0000011
					-- 0001100	0001000	0000100	0000001
					-- 0001001	0000101	0000010	0000000
					
					-- TRANSFORM BLOCK = { 
					--	  0							 1						  2						   3
					--0 { Coeffsin(111 downto 105) , Coeffsin(83 downto 77) , Coeffsin(55 downto 49) , Coeffsin(27 downto 21) }
					--1 { Coeffsin(104 downto 98) ,	 Coeffsin(76 downto 70) , Coeffsin(48 downto 42) , Coeffsin(20 downto 14) }
					--2 { Coeffsin(97 downto 91) ,	 Coeffsin(69 downto 63) , Coeffsin(41 downto 35) , Coeffsin(13 downto 7)  }
					--3 { Coeffsin(90 downto 84) ,	 Coeffsin(62 downto 56) , Coeffsin(34 downto 28) , Coeffsin(6 downto 0)   }
					-- }
					
					-- Read inputs
					TRANSFORM_BLOCK(0,0) := Coeffsin(111 downto 105);
					TRANSFORM_BLOCK(0,1) := Coeffsin(104 downto 98);
					TRANSFORM_BLOCK(0,2) := Coeffsin(97 downto 91);
					TRANSFORM_BLOCK(0,3) := Coeffsin(90 downto 84);
					TRANSFORM_BLOCK(1,0) := Coeffsin(83 downto 77);
					TRANSFORM_BLOCK(1,1) := Coeffsin(76 downto 70);
					TRANSFORM_BLOCK(1,2) := Coeffsin(69 downto 63);
					TRANSFORM_BLOCK(1,3) := Coeffsin(62 downto 56);
					TRANSFORM_BLOCK(2,0) := Coeffsin(55 downto 49);
					TRANSFORM_BLOCK(2,1) := Coeffsin(48 downto 42);
					TRANSFORM_BLOCK(2,2) := Coeffsin(41 downto 35);
					TRANSFORM_BLOCK(2,3) := Coeffsin(34 downto 28);
					TRANSFORM_BLOCK(3,0) := Coeffsin(27 downto 21);
					TRANSFORM_BLOCK(3,1) := Coeffsin(20 downto 14);
					TRANSFORM_BLOCK(3,2) := Coeffsin(13 downto 7);
					TRANSFORM_BLOCK(3,3) := Coeffsin(6 downto 0);
					
										-- -- DIAGONAL SCAN
					-- sig_coeff_flag(15)	:=	Coeffsin(0)		and	Coeffsin(1)		and	Coeffsin(2)		and	Coeffsin(3)		and	Coeffsin(4)		and	Coeffsin(5)		and	Coeffsin(6);
					-- sig_coeff_flag(14)	:=	Coeffsin(7)		and	Coeffsin(8)		and	Coeffsin(9)		and	Coeffsin(10)	and	Coeffsin(11)	and	Coeffsin(12)	and	Coeffsin(13);
					-- sig_coeff_flag(12)	:=	Coeffsin(14)	and	Coeffsin(15)	and	Coeffsin(16)	and	Coeffsin(17)	and	Coeffsin(18)	and	Coeffsin(19)	and	Coeffsin(20);
					-- sig_coeff_flag(9)	:=	Coeffsin(21)	and	Coeffsin(22)	and	Coeffsin(23)	and	Coeffsin(24)	and	Coeffsin(25)	and	Coeffsin(26)	and	Coeffsin(27);
					-- sig_coeff_flag(13)	:=	Coeffsin(28)	and	Coeffsin(29)	and	Coeffsin(30)	and	Coeffsin(31)	and	Coeffsin(32)	and	Coeffsin(33)	and	Coeffsin(34);
					-- sig_coeff_flag(11)	:=	Coeffsin(35)	and	Coeffsin(36)	and	Coeffsin(37)	and	Coeffsin(38)	and	Coeffsin(39)	and	Coeffsin(40)	and	Coeffsin(41);
					-- sig_coeff_flag(8)	:=	Coeffsin(42)	and	Coeffsin(43)	and	Coeffsin(44)	and	Coeffsin(45)	and	Coeffsin(46)	and	Coeffsin(47)	and	Coeffsin(48);
					-- sig_coeff_flag(5)	:=	Coeffsin(49)	and	Coeffsin(50)	and	Coeffsin(51)	and	Coeffsin(52)	and	Coeffsin(53)	and	Coeffsin(54)	and	Coeffsin(55);
					-- sig_coeff_flag(10)	:=	Coeffsin(56)	and	Coeffsin(57)	and	Coeffsin(58)	and	Coeffsin(59)	and	Coeffsin(60)	and	Coeffsin(61)	and	Coeffsin(62);
					-- sig_coeff_flag(7)	:=	Coeffsin(63)	and	Coeffsin(64)	and	Coeffsin(65)	and	Coeffsin(66)	and	Coeffsin(67)	and	Coeffsin(68)	and	Coeffsin(69);
					-- sig_coeff_flag(4)	:=	Coeffsin(70)	and	Coeffsin(71)	and	Coeffsin(72)	and	Coeffsin(73)	and	Coeffsin(74)	and	Coeffsin(75)	and	Coeffsin(76);
					-- sig_coeff_flag(2)	:=	Coeffsin(77)	and	Coeffsin(78)	and	Coeffsin(79)	and	Coeffsin(80)	and	Coeffsin(81)	and	Coeffsin(82)	and	Coeffsin(83);
					-- sig_coeff_flag(6)	:=	Coeffsin(84)	and	Coeffsin(85)	and	Coeffsin(86)	and	Coeffsin(87)	and	Coeffsin(88)	and	Coeffsin(89)	and	Coeffsin(90);
					-- sig_coeff_flag(3)	:=	Coeffsin(91)	and	Coeffsin(92)	and	Coeffsin(93)	and	Coeffsin(94)	and	Coeffsin(95)	and	Coeffsin(96)	and	Coeffsin(97);
					-- sig_coeff_flag(1)	:=	Coeffsin(98)	and	Coeffsin(99)	and	Coeffsin(100)	and	Coeffsin(101)	and	Coeffsin(102)	and	Coeffsin(103)	and	Coeffsin(104);
					-- sig_coeff_flag(0)	:=	Coeffsin(105)	and	Coeffsin(106)	and	Coeffsin(107)	and	Coeffsin(108)	and	Coeffsin(109)	and	Coeffsin(110)	and	Coeffsin(111);
					
					
					-- DIAGONAL SCAN
					sig_coeff_flag(0)	:=	Coeffsin(0)		or	Coeffsin(1)		or	Coeffsin(2)		or	Coeffsin(3)		or	Coeffsin(4)		or	Coeffsin(5)		or	Coeffsin(6);
					sig_coeff_flag(1)	:=	Coeffsin(7)		or	Coeffsin(8)		or	Coeffsin(9)		or	Coeffsin(10)	or	Coeffsin(11)	or	Coeffsin(12)	or	Coeffsin(13);
					sig_coeff_flag(3)	:=	Coeffsin(14)	or	Coeffsin(15)	or	Coeffsin(16)	or	Coeffsin(17)	or	Coeffsin(18)	or	Coeffsin(19)	or	Coeffsin(20);
					sig_coeff_flag(6)	:=	Coeffsin(21)	or	Coeffsin(22)	or	Coeffsin(23)	or	Coeffsin(24)	or	Coeffsin(25)	or	Coeffsin(26)	or	Coeffsin(27);
					sig_coeff_flag(2)	:=	Coeffsin(28)	or	Coeffsin(29)	or	Coeffsin(30)	or	Coeffsin(31)	or	Coeffsin(32)	or	Coeffsin(33)	or	Coeffsin(34);
					sig_coeff_flag(4)	:=	Coeffsin(35)	or	Coeffsin(36)	or	Coeffsin(37)	or	Coeffsin(38)	or	Coeffsin(39)	or	Coeffsin(40)	or	Coeffsin(41);
					sig_coeff_flag(7)	:=	Coeffsin(42)	or	Coeffsin(43)	or	Coeffsin(44)	or	Coeffsin(45)	or	Coeffsin(46)	or	Coeffsin(47)	or	Coeffsin(48);
					sig_coeff_flag(10)	:=	Coeffsin(49)	or	Coeffsin(50)	or	Coeffsin(51)	or	Coeffsin(52)	or	Coeffsin(53)	or	Coeffsin(54)	or	Coeffsin(55);
					sig_coeff_flag(5)	:=	Coeffsin(56)	or	Coeffsin(57)	or	Coeffsin(58)	or	Coeffsin(59)	or	Coeffsin(60)	or	Coeffsin(61)	or	Coeffsin(62);
					sig_coeff_flag(8)	:=	Coeffsin(63)	or	Coeffsin(64)	or	Coeffsin(65)	or	Coeffsin(66)	or	Coeffsin(67)	or	Coeffsin(68)	or	Coeffsin(69);
					sig_coeff_flag(11)	:=	Coeffsin(70)	or	Coeffsin(71)	or	Coeffsin(72)	or	Coeffsin(73)	or	Coeffsin(74)	or	Coeffsin(75)	or	Coeffsin(76);
					sig_coeff_flag(13)	:=	Coeffsin(77)	or	Coeffsin(78)	or	Coeffsin(79)	or	Coeffsin(80)	or	Coeffsin(81)	or	Coeffsin(82)	or	Coeffsin(83);
					sig_coeff_flag(9)	:=	Coeffsin(84)	or	Coeffsin(85)	or	Coeffsin(86)	or	Coeffsin(87)	or	Coeffsin(88)	or	Coeffsin(89)	or	Coeffsin(90);
					sig_coeff_flag(12)	:=	Coeffsin(91)	or	Coeffsin(92)	or	Coeffsin(93)	or	Coeffsin(94)	or	Coeffsin(95)	or	Coeffsin(96)	or	Coeffsin(97);
					sig_coeff_flag(14)	:=	Coeffsin(98)	or	Coeffsin(99)	or	Coeffsin(100)	or	Coeffsin(101)	or	Coeffsin(102)	or	Coeffsin(103)	or	Coeffsin(104);
					sig_coeff_flag(15)	:=	Coeffsin(105)	or	Coeffsin(106)	or	Coeffsin(107)	or	Coeffsin(108)	or	Coeffsin(109)	or	Coeffsin(110)	or	Coeffsin(111);
					
					
					sig_abs_level_greater1_flag(0)	:=	Coeffsin(1)		or	Coeffsin(2)		or	Coeffsin(3)		or	Coeffsin(4)		or	Coeffsin(5)		or	Coeffsin(6);
					sig_abs_level_greater1_flag(1)	:=	Coeffsin(8)		or	Coeffsin(9)		or	Coeffsin(10)	or	Coeffsin(11)	or	Coeffsin(12)	or	Coeffsin(13);
					sig_abs_level_greater1_flag(3)	:=  Coeffsin(15)	or	Coeffsin(16)	or	Coeffsin(17)	or	Coeffsin(18)	or	Coeffsin(19)	or	Coeffsin(20);
					sig_abs_level_greater1_flag(6)	:=	Coeffsin(22)	or	Coeffsin(23)	or	Coeffsin(24)	or	Coeffsin(25)	or	Coeffsin(26)	or	Coeffsin(27);
					sig_abs_level_greater1_flag(2)	:=	Coeffsin(29)	or	Coeffsin(30)	or	Coeffsin(31)	or	Coeffsin(32)	or	Coeffsin(33)	or	Coeffsin(34);
					sig_abs_level_greater1_flag(4)	:=	Coeffsin(36)	or	Coeffsin(37)	or	Coeffsin(38)	or	Coeffsin(39)	or	Coeffsin(40)	or	Coeffsin(41);
					sig_abs_level_greater1_flag(7)	:=	Coeffsin(43)	or	Coeffsin(44)	or	Coeffsin(45)	or	Coeffsin(46)	or	Coeffsin(47)	or	Coeffsin(48);
					sig_abs_level_greater1_flag(10)	:=	Coeffsin(50)	or	Coeffsin(51)	or	Coeffsin(52)	or	Coeffsin(53)	or	Coeffsin(54)	or	Coeffsin(55);
					sig_abs_level_greater1_flag(5)	:=	Coeffsin(57)	or	Coeffsin(58)	or	Coeffsin(59)	or	Coeffsin(60)	or	Coeffsin(61)	or	Coeffsin(62);
					sig_abs_level_greater1_flag(8)	:=	Coeffsin(64)	or	Coeffsin(65)	or	Coeffsin(66)	or	Coeffsin(67)	or	Coeffsin(68)	or	Coeffsin(69);
					sig_abs_level_greater1_flag(11)	:=	Coeffsin(71)	or	Coeffsin(72)	or	Coeffsin(73)	or	Coeffsin(74)	or	Coeffsin(75)	or	Coeffsin(76);
					sig_abs_level_greater1_flag(13)	:=	Coeffsin(78)	or	Coeffsin(79)	or	Coeffsin(80)	or	Coeffsin(81)	or	Coeffsin(82)	or	Coeffsin(83);
					sig_abs_level_greater1_flag(9)	:=	Coeffsin(85)	or	Coeffsin(86)	or	Coeffsin(87)	or	Coeffsin(88)	or	Coeffsin(89)	or	Coeffsin(90);
					sig_abs_level_greater1_flag(12)	:=	Coeffsin(92)	or	Coeffsin(93)	or	Coeffsin(94)	or	Coeffsin(95)	or	Coeffsin(96)	or	Coeffsin(97);
					sig_abs_level_greater1_flag(14)	:=	Coeffsin(99)	or	Coeffsin(100)	or	Coeffsin(101)	or	Coeffsin(102)	or	Coeffsin(103)	or	Coeffsin(104);
					sig_abs_level_greater1_flag(15)	:=	Coeffsin(106)	or	Coeffsin(107)	or	Coeffsin(108)	or	Coeffsin(109)	or	Coeffsin(110)	or	Coeffsin(111);
					
					
					sig_abs_level_greater2_flag(0)	:=	Coeffsin(2)		or	Coeffsin(3)		or	Coeffsin(4)		or	Coeffsin(5)		or	Coeffsin(6);
					sig_abs_level_greater2_flag(1)	:=	Coeffsin(9)		or	Coeffsin(10)	or	Coeffsin(11)	or	Coeffsin(12)	or	Coeffsin(13);
					sig_abs_level_greater2_flag(3)	:=  Coeffsin(16)	or	Coeffsin(17)	or	Coeffsin(18)	or	Coeffsin(19)	or	Coeffsin(20);
					sig_abs_level_greater2_flag(6)	:=	Coeffsin(23)	or	Coeffsin(24)	or	Coeffsin(25)	or	Coeffsin(26)	or	Coeffsin(27);
					sig_abs_level_greater2_flag(2)	:=	Coeffsin(30)	or	Coeffsin(31)	or	Coeffsin(32)	or	Coeffsin(33)	or	Coeffsin(34);
					sig_abs_level_greater2_flag(4)	:=	Coeffsin(37)	or	Coeffsin(38)	or	Coeffsin(39)	or	Coeffsin(40)	or	Coeffsin(41);
					sig_abs_level_greater2_flag(7)	:=	Coeffsin(44)	or	Coeffsin(45)	or	Coeffsin(46)	or	Coeffsin(47)	or	Coeffsin(48);
					sig_abs_level_greater2_flag(10)	:=	Coeffsin(51)	or	Coeffsin(52)	or	Coeffsin(53)	or	Coeffsin(54)	or	Coeffsin(55);
					sig_abs_level_greater2_flag(5)	:=	Coeffsin(58)	or	Coeffsin(59)	or	Coeffsin(60)	or	Coeffsin(61)	or	Coeffsin(62);
					sig_abs_level_greater2_flag(8)	:=	Coeffsin(65)	or	Coeffsin(66)	or	Coeffsin(67)	or	Coeffsin(68)	or	Coeffsin(69);
					sig_abs_level_greater2_flag(11)	:=	Coeffsin(72)	or	Coeffsin(73)	or	Coeffsin(74)	or	Coeffsin(75)	or	Coeffsin(76);
					sig_abs_level_greater2_flag(13)	:=	Coeffsin(79)	or	Coeffsin(80)	or	Coeffsin(81)	or	Coeffsin(82)	or	Coeffsin(83);
					sig_abs_level_greater2_flag(9)	:=	Coeffsin(86)	or	Coeffsin(87)	or	Coeffsin(88)	or	Coeffsin(89)	or	Coeffsin(90);
					sig_abs_level_greater2_flag(12)	:=	Coeffsin(93)	or	Coeffsin(94)	or	Coeffsin(95)	or	Coeffsin(96)	or	Coeffsin(97);
					sig_abs_level_greater2_flag(14)	:=	Coeffsin(100)	or	Coeffsin(101)	or	Coeffsin(102)	or	Coeffsin(103)	or	Coeffsin(104);
					sig_abs_level_greater2_flag(15)	:=	Coeffsin(107)	or	Coeffsin(108)	or	Coeffsin(109)	or	Coeffsin(110)	or	Coeffsin(111);


					--report "Coeffsin(111 downto 0) =" & std_logic'image(Coeffsin(0))& std_logic'image(Coeffsin(1))& std_logic'image(Coeffsin(2))& std_logic'image(Coeffsin(3))& std_logic'image(Coeffsin(4))& std_logic'image(Coeffsin(5))& std_logic'image(Coeffsin(6))& std_logic'image(Coeffsin(7))& std_logic'image(Coeffsin(8))& std_logic'image(Coeffsin(9))& std_logic'image(Coeffsin(10))& std_logic'image(Coeffsin(11))& std_logic'image(Coeffsin(12))& std_logic'image(Coeffsin(13))& std_logic'image(Coeffsin(14))& std_logic'image(Coeffsin(15))& std_logic'image(Coeffsin(16))& std_logic'image(Coeffsin(17))& std_logic'image(Coeffsin(18))& std_logic'image(Coeffsin(19))& std_logic'image(Coeffsin(20))& std_logic'image(Coeffsin(21))& std_logic'image(Coeffsin(22))& std_logic'image(Coeffsin(23))& std_logic'image(Coeffsin(24))& std_logic'image(Coeffsin(25))& std_logic'image(Coeffsin(26))& std_logic'image(Coeffsin(27))& std_logic'image(Coeffsin(28))& std_logic'image(Coeffsin(29))& std_logic'image(Coeffsin(30))& std_logic'image(Coeffsin(31))& std_logic'image(Coeffsin(32))& std_logic'image(Coeffsin(33))& std_logic'image(Coeffsin(34))& std_logic'image(Coeffsin(35))& std_logic'image(Coeffsin(36))& std_logic'image(Coeffsin(37))& std_logic'image(Coeffsin(38))& std_logic'image(Coeffsin(39))& std_logic'image(Coeffsin(40))& std_logic'image(Coeffsin(41))& std_logic'image(Coeffsin(42))& std_logic'image(Coeffsin(43))& std_logic'image(Coeffsin(44))& std_logic'image(Coeffsin(45))& std_logic'image(Coeffsin(46))& std_logic'image(Coeffsin(47))& std_logic'image(Coeffsin(48))& std_logic'image(Coeffsin(49))& std_logic'image(Coeffsin(50))& std_logic'image(Coeffsin(51))& std_logic'image(Coeffsin(52))& std_logic'image(Coeffsin(53))& std_logic'image(Coeffsin(54))& std_logic'image(Coeffsin(55))& std_logic'image(Coeffsin(56))& std_logic'image(Coeffsin(57))& std_logic'image(Coeffsin(58))& std_logic'image(Coeffsin(59))& std_logic'image(Coeffsin(60))& std_logic'image(Coeffsin(61))& std_logic'image(Coeffsin(62))& std_logic'image(Coeffsin(63))& std_logic'image(Coeffsin(64))& std_logic'image(Coeffsin(65))& std_logic'image(Coeffsin(66))& std_logic'image(Coeffsin(67))& std_logic'image(Coeffsin(68))& std_logic'image(Coeffsin(69))& std_logic'image(Coeffsin(70))& std_logic'image(Coeffsin(71))& std_logic'image(Coeffsin(72))& std_logic'image(Coeffsin(73))& std_logic'image(Coeffsin(74))& std_logic'image(Coeffsin(75))& std_logic'image(Coeffsin(76))& std_logic'image(Coeffsin(77))& std_logic'image(Coeffsin(78))& std_logic'image(Coeffsin(79))& std_logic'image(Coeffsin(80))& std_logic'image(Coeffsin(81))& std_logic'image(Coeffsin(82))& std_logic'image(Coeffsin(83))& std_logic'image(Coeffsin(84))& std_logic'image(Coeffsin(85))& std_logic'image(Coeffsin(86))& std_logic'image(Coeffsin(87))& std_logic'image(Coeffsin(88))& std_logic'image(Coeffsin(89))& std_logic'image(Coeffsin(90))& std_logic'image(Coeffsin(91))& std_logic'image(Coeffsin(92))& std_logic'image(Coeffsin(93))& std_logic'image(Coeffsin(94))& std_logic'image(Coeffsin(95))& std_logic'image(Coeffsin(96))& std_logic'image(Coeffsin(97))& std_logic'image(Coeffsin(98))& std_logic'image(Coeffsin(99))& std_logic'image(Coeffsin(100))& std_logic'image(Coeffsin(101))& std_logic'image(Coeffsin(102))& std_logic'image(Coeffsin(103))& std_logic'image(Coeffsin(104))& std_logic'image(Coeffsin(105))& std_logic'image(Coeffsin(106))& std_logic'image(Coeffsin(107))& std_logic'image(Coeffsin(108))& std_logic'image(Coeffsin(109))& std_logic'image(Coeffsin(110))& std_logic'image(Coeffsin(111));
					
				end if;
				
					--outputs for debugging
					-- Coeffsout(111 downto 105) <= TRANSFORM_BLOCK(0,0);
					-- Coeffsout(104 downto 98)  <= TRANSFORM_BLOCK(0,1);
					-- Coeffsout(97 downto 91)   <= TRANSFORM_BLOCK(0,2);
					-- Coeffsout(90 downto 84)   <= TRANSFORM_BLOCK(0,3);
					-- Coeffsout(83 downto 77)   <= TRANSFORM_BLOCK(1,0);
					-- Coeffsout(76 downto 70)   <= TRANSFORM_BLOCK(1,1);
					-- Coeffsout(69 downto 63)   <= TRANSFORM_BLOCK(1,2);
					-- Coeffsout(62 downto 56)   <= TRANSFORM_BLOCK(1,3);
					-- Coeffsout(55 downto 49)   <= TRANSFORM_BLOCK(2,0);
					-- Coeffsout(48 downto 42)   <= TRANSFORM_BLOCK(2,1);
					-- Coeffsout(41 downto 35)   <= TRANSFORM_BLOCK(2,2);
					-- Coeffsout(34 downto 28)   <= TRANSFORM_BLOCK(2,3);
					-- Coeffsout(27 downto 21)   <= TRANSFORM_BLOCK(3,0);
					-- Coeffsout(20 downto 14)   <= TRANSFORM_BLOCK(3,1);
					-- Coeffsout(13 downto 7)    <= TRANSFORM_BLOCK(3,2);
					-- Coeffsout(6 downto 0)     <= TRANSFORM_BLOCK(3,3);
				
			end if;
		
		end if;
	
	end if;

    end process;

end arch_imp;
